/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.util.Set;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.commons.threadlocal.DocumentBuilders;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * 
 * @author alouis
 * 
 */
public class PatternHelper {

    private static final XPath XPATH;
    static {
        XPATH = XPathFactory.newInstance().newXPath();
    }

    /**
     * Process the result of a sent exchange.<br>
     * - if the exchange contains an OUT, return it and close the sent exchange<br>
     * - if the exchange contains a FAULT, return it and close the sent exchange<br>
     * - if the exchange is on ERROR status, throw a JBI Exception - if the
     * exchange is on DONE status, return a null Normalized Message<br>
     * 
     * @param sentExchange
     * @param originalExchange
     * @param context
     * @param setFailure
     * @return the OUT message of the sent exchange, if any. When null, the
     *         original exchange has been updated
     * @throws JBIException
     *             An exception from the sentExchange
     */
    public static final NormalizedMessage processSentExchange(final Exchange sentExchange,
            final ExchangeContext context) throws JBIException {
        NormalizedMessage receivedMessage = null;

        if (sentExchange.isActiveStatus()) {
            if (sentExchange.getFault() != null) {
                // fault received, return it
                receivedMessage = sentExchange.getFault();
            } else {
                // out received, return it
                receivedMessage = sentExchange.getOutMessage();
            }
            // close the sent exchange
            sentExchange.setStatus(ExchangeStatus.DONE);
            context.send(sentExchange);
        } else if (sentExchange.isErrorStatus()) {
            // error on the sent exchange
            if (sentExchange.getError() == null) {
                throw new JBIException("An unknown error occurs");
            } else {
                throw new JBIException(sentExchange.getError());
            }
        }
        return receivedMessage;
    }

    /**
     * Copies message's content, security, properties and attachments in
     * another.
     * 
     * @param source
     *            The NormalizedMessage source (copied)
     * @param target
     *            The NormalizedMessage target (pasted)
     * @throws MessagingException
     *             error while setting attachment
     * @throws PEtALSCDKException
     *             error while using PEtALS object
     */
    public static final void copy(final NormalizedMessage source, final NormalizedMessage target)
    throws MessagingException, PEtALSCDKException {
        copy(source, target, true, true, true, true);
    }

    /**
     * Copies message's properties, attachments, content and security provided,
     * in another.
     * 
     * @param source
     * @param target
     * @param copyContent
     *            Copy the content or not
     * @param copySecurity
     *            Copy the security or not
     * @param copyAttachments
     *            Copy the attachments or not
     * @param copyProperties
     *            Copy the properties or not
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    public static final void copy(final NormalizedMessage source, final NormalizedMessage target,
            final boolean copyContent, final boolean copySecurity, final boolean copyAttachments,
            final boolean copyProperties) throws MessagingException, PEtALSCDKException {
        if (source != null) {
            // setting content
            if (copyContent == true) {
                target.setContent(source.getContent());
            }
            if (copySecurity == true) {
                // setting security context
                if (source.getSecuritySubject() != null) {
                    target.setSecuritySubject(source.getSecuritySubject());
                }
            }
            // setting attachments
            if (copyAttachments == true) {
                copyAttachments(source, target);
            }
            // setting properties
            if (copyProperties == true) {
                copyProperties(source, target);
            }
        }
    }

    /**
     * Retrieve the MEP from the extensions.
     * 
     * @param extensions
     * @return mep specified in extensions
     * @throws MessagingException
     *             MEP pattern specified in consumes extension isn't valid or if
     *             no MEP pattern specified in extensions.
     */
    public static final MEPConstants retrievePattern(final MEPType patternShortcut)
    throws MessagingException {
        MEPConstants mep = null;

        if (patternShortcut != null) {
            if (MEPType.IN_ONLY.equals(patternShortcut)) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else if (MEPType.ROBUST_IN_ONLY.equals(patternShortcut)) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (MEPType.IN_OUT.equals(patternShortcut)) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (MEPType.IN_OPTIONAL_OUT.equals(patternShortcut)) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            }
        }

        return mep;
    }

    /**
     * 
     * @param condition
     * @param message
     * @param context
     * @return
     * @throws XPathExpressionException
     */
    public static final boolean evaluateXPathCondition(final String expression, final Node message)
    throws XPathExpressionException {
        boolean evaluation = false;
        final String result = evaluateXPathExpressionAsString(expression, message);
        if (result != null) {
            evaluation = result.equals(Boolean.TRUE.toString());
        }
        return evaluation;
    }

    /**
     * Evaluate a XPath expression
     * 
     * @param expression
     *            The XPath expression to compile
     * @param message
     *            The Node to evaluate
     * @return The String evaluated by XPath under the message
     * @throws XPathExpressionException
     */
    public static final String evaluateXPathExpressionAsString(final String expression,
            final Node message) throws XPathExpressionException {
        final XPathExpression xpathExpression = XPATH.compile(expression);
        final String result = xpathExpression.evaluate(message);
        XPATH.reset();
        return result;
    }

    /**
     * Evaluate a XPath expression
     * 
     * @param expression
     *            The XPath expression to compile
     * @param message
     *            The Node to evaluate
     * @return The NodeList evaluated by XPath under the message
     * @throws XPathExpressionException
     */
    public static final NodeList evaluateXPathExpressionAsNodeList(final String expression,
            final Node message) throws XPathExpressionException {
        NodeList nodeList = null;
        final XPathExpression xpathExpression = XPATH.compile(expression);
        nodeList = (NodeList) xpathExpression.evaluate(message, XPathConstants.NODESET);
        XPATH.reset();
        return nodeList;
    }

    /**
     * Handle the streaming issue.
     * 
     * @param streamSource
     * @return a forker or null if the content of the message isn't a stream
     *         source
     * @throws MessagingException
     */
    public static final InputStreamForker handleStreamSource(final NormalizedMessage message)
    throws MessagingException {
        InputStreamForker inMessageAsStream = null;
        final Source content = message.getContent();
        if (content instanceof StreamSource) {
            inMessageAsStream = new InputStreamForker(((StreamSource) content).getInputStream());
            final Source inSource = new StreamSource(inMessageAsStream.getInputStreamOne());
            message.setContent(inSource);
        }
        return inMessageAsStream;
    }

    /**
     * Handle the streaming issue.
     * 
     * @param streamSource
     * @throws MessagingException
     */
    public static final InputStreamForker handleStreamSource(final Fault fault)
    throws MessagingException {
        InputStreamForker faultMessageAsStream = null;
        if (fault.getContent() instanceof StreamSource) {
            faultMessageAsStream = new InputStreamForker(((StreamSource) fault.getContent())
                    .getInputStream());
            final Source faultSource = new StreamSource(faultMessageAsStream.getInputStreamOne());
            fault.setContent(faultSource);
        }

        return faultMessageAsStream;
    }

    /**
     * Copy the attachments.
     * 
     * @param sourceMessage
     *            message containing the attachments
     * @param destMessage
     * 
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    @SuppressWarnings("unchecked")
    public static void copyAttachments(final NormalizedMessage sourceMessage,
            final NormalizedMessage destMessage) throws PEtALSCDKException, MessagingException {
        final Set<String> attNames = sourceMessage.getAttachmentNames();
        for (final String string : attNames) {
            destMessage.addAttachment(string, sourceMessage.getAttachment(string));
        }
    }

    /**
     * Copy the properties.
     * 
     * @param destMessage
     *            message containing the properties
     * @param destMessage
     * 
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    @SuppressWarnings("unchecked")
    public static void copyProperties(final NormalizedMessage sourceMessage,
            final NormalizedMessage destMessage) throws PEtALSCDKException, MessagingException {
        final Set<String> attNames = sourceMessage.getPropertyNames();
        for (final String string : attNames) {
            destMessage.setProperty(string, sourceMessage.getProperty(string));
        }
    }

    /**
     * Create an document to fill with response nodes.
     * 
     * @param namespace
     *            The namespace of the root element
     * @return the {@code document}
     */
    public static Document createResultDocument(final String namespace) {
        return createDocument(EIPConstants.RESULT_ELEMENT, namespace);
    }

    /**
     * Create an document for the exceptions.
     * 
     * @param namespace
     *            The namespace of the root element
     * @return the {@code document}
     */
    public static Document createExceptionDocument(final String namespace) {
        return createDocument(EIPConstants.EXCEPTION_ELEMENT, namespace);
    }

    /**
     * Create an document
     * 
     * @param name
     *            The name of the root element
     * @param namespace
     *            The namespace of the root element
     * @return the {@code document}
     */
    private static Document createDocument(final String name, final String namespace) {
        // build an aggregate document for the result
        final Document aggregateDocument = DocumentBuilders.getNamespaceDocumentBuilder()
        .newDocument();
        final Node aggregateRootElement = aggregateDocument.createElementNS(namespace, name);
        aggregateDocument.appendChild(aggregateRootElement);
        return aggregateDocument;
    }

    /**
     * Aggregate the item XML content into the global AggregateDocument
     * 
     * @param document
     *            xml content to append
     * @param aggregateDocument
     *            the global document in which the item is appended
     * @param namespace
     *            The namespace of the root element
     * @return the document appended
     * @throws JBIException
     * @throws PEtALSCDKException
     */
    public static Document aggregateContent(final Document document, Document aggregatedDocument,
            final String contentName, final String namespace) throws JBIException {
        Node contentNode = null;
        Element itemElement = null;

        if (aggregatedDocument == null) {
            aggregatedDocument = createResultDocument(namespace);
        }
        // adopt the document
        itemElement = document.getDocumentElement();
        if (itemElement != null) {
            aggregatedDocument.adoptNode(itemElement);

            final NodeList nodeList = aggregatedDocument.getElementsByTagNameNS(namespace,
                    contentName);
            if (nodeList == null || nodeList.getLength() == 0) {
                aggregatedDocument = createResultDocument(namespace);
            } else {
                boolean found = false;
                for (int i = 0; i < nodeList.getLength() && !found; i++) {
                    if (aggregatedDocument.getDocumentElement().isSameNode(nodeList.item(i))) {
                        found = true;
                        contentNode = nodeList.item(i);
                    }
                    if (contentName == null) {
                        throw new JBIException("unable to found the root node to aggregate");
                    }
                    contentNode.appendChild(itemElement);
                }
            }
        }
        return aggregatedDocument;
    }

    /**
     * Aggregate the received normalized message content into the global
     * aggregated document
     * 
     * @param receivedMessage
     *            message containing the xml content to append
     * @param aggregatedDocument
     *            the global document in which the content is appended
     * @param name
     *            The name of the new element
     * @param namespace
     *            The namespace of the new element
     * @return The aggregated document
     * @throws PEtALSCDKException
     */
    public static Document aggregateContent(final NormalizedMessage receivedMessage,
            final Document aggregatedDocument, final String name, final String namespace)
    throws JBIException {
        final Source source = receivedMessage.getContent();
        if (source != null) {
            final Document itemDocument = UtilFactory.getSourceUtil().createDocument(source);
            if (itemDocument != null) {
                return aggregateContent(itemDocument, aggregatedDocument, name, namespace);
            }
        }
        return aggregatedDocument;
    }
}
